$( "#{{ name }}" ).click(function() {

    {% if submit_first %}
        var data = $("#form").serializeArray();
        $.post("/callable/set_page_data", data);
    {% endif %}

    $.get( "{{ url }}", function(data) {
    
        if ("{{ followup }}" == "refresh") {
            move(direction="stay");
        } else if ("{{ followup }}" == "custom") {
            {{ custom_js }}
        } else if ("{{ followup }}" != "none") {
            move(direction="{{ followup }}");
        }
    });
})
