<div class="{{ css_class_element }}" id="{{ name }}">
{% for choice in choices %}
<div class="form-check {% if not vertical %} form-check-inline {% endif %} {{ align }}" style="{{ height }} {{ fontsize }}">
  <input class="form-check-input" type="{{ choice.type }}" value="{{ choice.value }}" id="{{ choice.id }}"
    name="{{ choice.name }}" {% if choice.checked %}checked{% endif %} {% if choice.disabled %}disabled{% endif %}>
  <label class="form-check-label" for="{{ choice.id }}">
    {{ choice.label }}
  </label>
</div>
{% endfor %}
</div>
